# -*- coding: utf-8 -*-

import os
import sys

from importlib import metadata

WINDOWS = os.name == "nt"
DOCKER = os.path.isfile("/.dockerenv")
KUBERNETES = os.environ.get("KUBERNETES_SERVICE_HOST") is not None
EXECUTABLE = bool(getattr(sys, "frozen", False))
MEIPASS_PATH: str | None = getattr(sys, "_MEIPASS", None)
PYTHON_VERSION = sys.version_info

CONTAINER = DOCKER or KUBERNETES
MEIPASS = MEIPASS_PATH is not None
PYINSTALLER = EXECUTABLE and MEIPASS


def resource_path(relative_path: str):
    """Return absolute path to resource for frozen PyInstaller executable."""
    if PYINSTALLER:
        assert MEIPASS_PATH
        return os.path.join(MEIPASS_PATH, get_package_name(), relative_path)
    else:
        return os.path.join(os.path.dirname(os.path.abspath(__file__)), relative_path)


def get_log_file_path(log_dir: str):
    """Get log file path depending on the package location."""
    is_installed = is_package_installed("gallery-dl-server")

    if log_dir or is_installed:
        filename = "gallery-dl-server.log"
    else:
        filename = "app.log"

    if not log_dir:
        if is_installed:
            log_dir = os.path.expanduser("~")
        else:
            log_dir = os.path.join(os.getcwd(), "logs")

    return os.path.join(log_dir, filename)


def dirname_parent(path: str):
    """Return grandparent directory of the given path."""
    return os.path.dirname(os.path.dirname(path))


def join_paths(base_path: str, /, *paths: str):
    """Join each path with the base path and return a list."""
    return [os.path.join(base_path, path) for path in paths]


def get_package_name():
    """Return the name of the package."""
    fallback = os.path.basename(os.path.dirname(os.path.abspath(__file__)))
    return __package__ if __package__ else fallback


def is_package(package_name: str):
    """Check if the package exists in the current working directory."""
    path = os.path.join(os.getcwd(), package_name)
    return os.path.isdir(path) and os.path.isfile(os.path.join(path, "__init__.py"))


def is_package_installed(installed_name: str):
    """Check if the package is installed in the current environment and not
    in the current working directory."""
    installed = {dist.metadata["Name"] for dist in metadata.distributions()}
    return installed_name in installed if not is_package(get_package_name()) else False


def normalise_path(path: str):
    """Expands user constructions and environment variables in the given path,
    normalises it, and returns the absolute path."""
    return (
        os.path.abspath(os.path.normpath(os.path.expandvars(os.path.expanduser(path))))
        if path
        else path
    )


def is_imported(module_name: str):
    """Check if a module has been imported."""
    return module_name in sys.modules


def filter_integers(values: list[int | str | None]):
    """Return only the integer values in a list."""
    return [value for value in values if isinstance(value, int)]



# -*- coding: utf-8 -*-

import logging
import os
import sys

from typing import Any

from gallery_dl import config

from . import output, utils

_config: dict[str, Any] = config._config
_files: list[str] = config._files

log = output.initialise_logging(__name__)


def clear(conf: dict[str, Any] = _config):
    """Clear loaded configuration."""
    conf.clear()


def get_default_configs():
    """Return default gallery-dl configuration file locations."""
    if utils.CONTAINER:
        _default_configs = [
            "/config/gallery-dl.conf",
            "/config/config.json",
        ]
    elif utils.WINDOWS:
        _default_configs = [
            "%APPDATA%\\gallery-dl\\config.json",
            "%USERPROFILE%\\gallery-dl\\config.json",
            "%USERPROFILE%\\gallery-dl.conf",
        ]
    else:
        _default_configs = [
            "/etc/gallery-dl.conf",
            "${XDG_CONFIG_HOME}/gallery-dl/config.json"
            if os.environ.get("XDG_CONFIG_HOME")
            else "${HOME}/.config/gallery-dl/config.json",
            "${HOME}/.gallery-dl.conf",
        ]

    if utils.EXECUTABLE:
        _default_configs.extend(
            utils.join_paths(
                os.path.dirname(sys.executable),
                "gallery-dl.conf",
                "config.json",
            )
        )

    return _default_configs


def get_new_configs(_configs: list[str], exts: list[str]):
    """Return list of original paths and paths with new extensions."""
    _new_configs: list[str] = []

    for path in _configs:
        _new_configs.append(path)

        for ext in exts:
            base_path = path.rsplit(".", 1)[0]
            _new_configs.append(base_path + ext)

    return _new_configs


def load():
    """Load gallery-dl configuration files."""
    configs_found: list[str] = []
    configs_loaded: list[str] = []
    exit_codes: list[int | str | None] = []
    messages: list[str] = []

    new_exts = [".toml", ".yaml", ".yml"]
    _configs = get_new_configs(get_default_configs(), new_exts)

    log_buffer = output.StringLogger()

    for path in _configs:
        normal_path = utils.normalise_path(path)

        if os.path.isfile(normal_path):
            configs_found.append(normal_path)
            log.info(f"Configuration file found: {normal_path}")

            try:
                load_config(path, exit_codes)
                configs_loaded.append(path)
            except ImportError as e:
                messages.append(f"{type(e).__name__}: {e}")
            except SystemExit as e:
                exit_codes.append(e.code)
                messages.append(log_buffer.get_logs().split(output.LOG_SEPARATOR)[-1])

    log_buffer.close()

    log_results(_configs, configs_loaded, configs_found, exit_codes, messages)


def load_config(path: str, exit_codes: list[int | str | None]):
    """Load a single configuration file based on its extension."""
    if path.endswith((".conf", ".json")):
        config.load([path], strict=True)
    elif path.endswith((".yaml", ".yml")):
        try:
            import yaml
        except ImportError:
            exit_codes.append(3)
            raise

        config.load([path], strict=True, loads=yaml.safe_load)
    elif path.endswith(".toml"):
        try:
            import tomllib as toml
        except ImportError:
            try:
                import toml  # type: ignore
            except ImportError:
                exit_codes.append(4)
                raise

        config.load([path], strict=True, loads=toml.loads)


def log_results(
    _configs: list[str],
    configs_loaded: list[str],
    configs_found: list[str],
    exit_codes: list[int | str | None],
    messages: list[str],
):
    """Log and handle the results of loading configuration files."""
    if configs_loaded:
        log.info(f"Loaded gallery-dl configuration file(s): [{output.join(_files)}]")
    elif all(code not in exit_codes for code in [2, 3, 4]):
        log.error("Loading configuration files failed with exit code: 1")

        if not configs_found:
            log.info(f"Valid configuration file locations: [{output.join(_configs)}]")
    elif all(code not in exit_codes for code in [3, 4]):
        log.error("Loading configuration files failed with exit code: 2")
    elif 2 not in exit_codes:
        log.error("Loading configuration files failed due to missing imports")

    for message in messages:
        log.log_multiline(logging.ERROR, message)

    if exit_codes:
        if 3 in exit_codes:
            log.info("Install 'PyYAML' to load YAML configuration files")
        if 4 in exit_codes:
            log.info("Install 'toml' or use Python >= 3.11 to load TOML configuration files")

        unique_exit_codes = sorted(set(utils.filter_integers(exit_codes)))
        if unique_exit_codes:
            log.debug(f"Exit codes: {unique_exit_codes}")

    if not configs_loaded:
        raise SystemExit(1)


def get(
    path: list[str],
    default: Any = None,
    conf: dict[str, Any] = _config,
):
    """Get value from nested dict or return default."""
    if isinstance(path, (list, tuple)):
        try:
            for p in path:
                conf = conf[p]
            return conf
        except Exception:
            return default


def add(
    _dict: dict[str, Any] | None = None,
    conf: dict[str, Any] = _config,
    fixed=False,
    **kwargs: Any,
):
    """Add entries to a nested dictionary or list."""
    if _dict:
        for k, v in _dict.items():
            if k in conf.keys() or not fixed:
                if isinstance(v, dict):
                    conf[k] = add(v, conf.get(k) or {}, fixed=fixed)[0]
                elif isinstance(v, list):
                    for i in v:
                        if not isinstance(i, dict):
                            if i not in conf.get(k, []) or not str(i).startswith("-"):
                                conf[k] = conf.get(k, []) + [i]
                        else:
                            if i not in conf.get(k, []):
                                conf[k] = conf.get(k, []) + [i]
                else:
                    conf[k] = v

        while isinstance(d := list(_dict.values())[0], dict):
            _dict = d

    if kwargs:
        for key, val in kwargs.items():
            for k, v in conf.items():
                if k == key and key in conf.keys():
                    conf[k] = val
                elif isinstance(v, dict):
                    conf[k] = add(conf=v, fixed=fixed, **{key: val})[0]

    return conf, [_dict] if not kwargs else [kwargs] if not _dict else [_dict, kwargs]


def remove(
    path: dict[str, Any] | list[Any],
    item: str | None = None,
    key: str | None = None,
    value: Any = None,
):
    """Remove entries from a nested dictionary or list."""
    entries_removed: list[Any] = []

    if isinstance(path, dict) and key:
        entries_removed.extend(remove_from_dict(path, key, value))
    elif isinstance(path, list) and (item or key):
        entries_removed.extend(remove_from_list(path, item, key, value))

    return entries_removed


def remove_from_dict(_dict: dict[str, Any], key: str, value: Any):
    """Remove keys from a nested dictionary."""
    keys_to_remove: list[str] = []
    keys_removed: list[dict[str, Any]] = []

    for k, v in _dict.items():
        if k == key and (value is None or v == value):
            keys_to_remove.append(k)

    for k in keys_to_remove:
        try:
            v = _dict.pop(k)
            keys_removed.append({k: v})
        except Exception as e:
            log.error(f"Exception: {type(e).__name__}", exc_info=True)

    return keys_removed


def remove_from_list(_list: list[Any], item: str | None, key: str | None, value: Any):
    """Remove elements from a nested list."""
    elements_to_remove: list[Any] = []
    elements_removed: list[Any] = []

    for element in _list:
        if isinstance(element, dict):
            if key and (element.get(key) == value or not value and element.get(key)):
                elements_to_remove.append(element)
        elif item and element == item:
            try:
                element_index = _list.index(element)
                element_next = _list[element_index + 1] if element_index + 1 < len(_list) else None
                if value == "any" or (element_next == value):
                    elements_to_remove.append(element)
                    if element_next:
                        elements_to_remove.append(element_next)
            except IndexError:
                if value == "any":
                    elements_to_remove.append(element)

    for element in elements_to_remove:
        try:
            _list.remove(element)
            elements_removed.append(element)
        except Exception as e:
            log.error(f"Exception: {type(e).__name__}", exc_info=True)

    return elements_removed



# -*- coding: utf-8 -*-

"""
gallery-dl-server: Web UI for downloading media with gallery-dl and yt-dlp.

This package serves as a middleware for gallery-dl and yt-dlp, providing a simple
web and REST interface for downloading media from various websites.
"""

import sys

if sys.version_info < (3, 10):
    raise ImportError(
        "You are using an unsupported version of Python. "
        "Please upgrade to Python 3.10 or above to use gallery-dl-server."
    )

from . import app, options, utils, version

__version__ = version.__version__
__all__ = ["run"]


def run(
    host: str = "0.0.0.0",
    port: int = 0,
    log_dir: str = "~",
    log_level: str = "info",
    server_log_level: str = "info",
    access_log: bool = False,
) -> None:
    """
    Run gallery-dl-server with custom options.

    Args:
        host (str): The host address for the server
            (`0.0.0.0` listens on all available interfaces, allowing access from any IP address).

        port (int): The port number for the server
            (`0` selects a random port).

        log_dir (str): The directory for the log file
            (defaults to the user's home directory on the operating system).

        log_level (str): The log level for downloads
            (accepted values: `critical`, `error`, `warning`, `info`, `debug`).

        server_log_level (str): The log level for the server
            (accepted values: `critical`, `error`, `warning`, `info`, `debug`, `trace`).

        access_log (bool): Enable or disable the access log only, without changing the log level
            (i.e. show `GET` requests, WebSocket connections, etc.).

    Raises:
        TypeError: If an invalid parameter is passed to the function, it will raise a `TypeError`.

    Examples:
        To run the server on `localhost` with a specific port:

        ```python
        import gallery_dl_server as server

        if __name__ == "__main__":
            server.run(host="127.0.0.1", port=9080, server_log_level="info", access_log=False)
        ```

        To run the server on all interfaces with a random port and enable access logging:

        ```python
        import gallery_dl_server as server

        if __name__ == "__main__":
            server.run(host="0.0.0.0", port=0, server_log_level="debug", access_log=True)
        ```

        The `if __name__ == "__main__"` guard is necessary on Windows to prevent the server from
        starting itself recursively when attempting to initiate a download.

        This is because the server runs each download in a child process using the
        `multiprocessing` module, which uses `spawn` as the default start method on Windows.

        The `spawn` method creates a new Python interpreter which imports the main module,
        causing any unguarded code to be executed again in the child process.

        See the following:
        - https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
        - https://docs.python.org/3/library/multiprocessing.html#the-spawn-and-forkserver-start-methods
    """
    kwargs = {
        "host": host,
        "port": port,
        "log_dir": utils.normalise_path(log_dir),
        "log_level": log_level.lower(),
        "server_log_level": server_log_level.lower(),
        "access_log": access_log,
    }

    try:
        args = options.custom_args = options.CustomNamespace(**kwargs)
        app.main(args=args)
    except TypeError:
        raise



# -*- coding: utf-8 -*-

import os

from argparse import ArgumentParser, Namespace

from . import utils

custom_args: "CustomNamespace | None" = None


def parse_args(is_main_module: bool = False):
    """Parse command-line arguments and return namespace with the correct types."""
    global custom_args
    if custom_args is not None:
        return custom_args

    prog_name = utils.get_package_name() if is_main_module else None

    parser = ArgumentParser(prog=prog_name, description="Run gallery-dl-server with custom options")

    parser.add_argument(
        "--host",
        type=str,
        default=os.environ.get("HOST", "0.0.0.0"),
        help="host address (default: 0.0.0.0)",
    )

    default_port = os.environ.get("PORT", "0")
    try:
        default_port = int(default_port)
    except ValueError:
        default_port = 0

    parser.add_argument(
        "--port",
        type=int,
        default=default_port,
        help="port number [0-65535] (default: any available port)",
    )

    parser.add_argument(
        "--log-dir",
        type=str,
        default=os.environ.get("LOG_DIR", ""),
        help="log file directory (default: user home directory)",
    )

    parser.add_argument(
        "--log-level",
        type=str,
        default=os.environ.get("LOG_LEVEL", "info"),
        help="download log level [critical|error|warning|info|debug] (default: info)",
    )

    parser.add_argument(
        "--server-log-level",
        type=str,
        default=os.environ.get("SERVER_LOG_LEVEL", "info"),
        help="server log level [critical|error|warning|info|debug|trace] (default: info)",
    )

    parser.add_argument(
        "--access-log",
        type=str,
        default=os.environ.get("ACCESS_LOG", "false"),
        help="enable server access logging [true|false] (default: false)",
    )

    args = parser.parse_args()

    custom_args = validate_args(parser, args)

    return custom_args


def validate_args(parser: ArgumentParser, args: Namespace):
    """Validate arguments and return the correct types."""
    host: str = args.host
    port: int = args.port
    log_dir: str = args.log_dir
    log_level: str = args.log_level
    server_log_level: str = args.server_log_level
    access_log: str = args.access_log

    if port < 0 or port > 65535:
        parser.error("invalid value for --port, must be a valid integer between 0 and 65535")

    if log_dir != "" and not os.path.isdir(utils.normalise_path(log_dir)):
        parser.error("invalid value for --log-dir, must be a path to an existing directory")

    log_levels = ["critical", "error", "warning", "info", "debug"]

    if log_level.lower() not in log_levels:
        parser.error("invalid value for --log-level, use --help to view the valid options")

    log_levels.append("trace")

    if server_log_level.lower() not in log_levels:
        parser.error("invalid value for --server-log-level, use --help to view the valid options")

    if access_log.lower() not in ["true", "false"]:
        parser.error("invalid value for --access-log, must be 'true' or 'false'")

    return CustomNamespace(
        host=host,
        port=port,
        log_dir=utils.normalise_path(log_dir),
        log_level=log_level.lower(),
        server_log_level=server_log_level.lower(),
        access_log=access_log.lower() == "true",
    )


def get_default_args():
    """Bypass argument parsing and return the default arguments."""
    host = os.environ.get("HOST", "0.0.0.0")
    port = os.environ.get("PORT", "0")
    log_dir = os.environ.get("LOG_DIR", "")
    log_level = os.environ.get("LOG_LEVEL", "info")
    server_log_level = os.environ.get("SERVER_LOG_LEVEL", "info")
    access_log = os.environ.get("ACCESS_LOG", "false")

    return CustomNamespace(
        host=host,
        port=int(port),
        log_dir=utils.normalise_path(log_dir),
        log_level=log_level.lower(),
        server_log_level=server_log_level.lower(),
        access_log=access_log.lower() == "true",
    )


class CustomNamespace(Namespace):
    """Custom namespace for type enforcement."""

    def __init__(
        self,
        host: str,
        port: int,
        log_dir: str,
        log_level: str,
        server_log_level: str,
        access_log: bool,
    ):
        super().__init__()
        self.host = host
        self.port = port
        self.log_dir = log_dir
        self.log_level = log_level
        self.server_log_level = server_log_level
        self.access_log = access_log

        self._validate_types()

    def _validate_types(self):
        """Perform type validation."""
        if not isinstance(self.host, str):
            raise TypeError(
                "Expected 'host' to be of type str, got {}".format(type(self.host).__name__)
            )

        if not isinstance(self.port, int):
            raise TypeError(
                "Expected 'port' to be of type int, got {}".format(type(self.port).__name__)
            )

        if not isinstance(self.log_dir, str):
            raise TypeError(
                "Expected 'log_dir' to be of type str, got {}".format(type(self.log_dir).__name__)
            )

        if not isinstance(self.log_level, str):
            raise TypeError(
                "Expected 'log_level' to be of type str, got {}".format(
                    type(self.log_level).__name__
                )
            )

        if not isinstance(self.server_log_level, str):
            raise TypeError(
                "Expected 'server_log_level' to be of type str, got {}".format(
                    type(self.server_log_level).__name__
                )
            )

        if not isinstance(self.access_log, bool):
            raise TypeError(
                "Expected 'access_log' to be of type bool, got {}".format(
                    type(self.access_log).__name__
                )
            )



# -*- coding: utf-8 -*-

import asyncio
import multiprocessing
import os
import queue
import shutil
import signal
import time

from contextlib import asynccontextmanager
from multiprocessing.queues import Queue
from types import FrameType
from typing import Any

import aiofiles
import watchfiles

from starlette.applications import Starlette
from starlette.background import BackgroundTask
from starlette.datastructures import UploadFile
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette.responses import RedirectResponse, JSONResponse, StreamingResponse
from starlette.requests import Request
from starlette.routing import Route, WebSocketRoute, Mount
from starlette.staticfiles import StaticFiles
from starlette.status import (
    HTTP_200_OK,
    HTTP_404_NOT_FOUND,
    HTTP_500_INTERNAL_SERVER_ERROR,
)
from starlette.templating import Jinja2Templates
from starlette.types import ASGIApp
from starlette.websockets import WebSocket, WebSocketDisconnect, WebSocketState

import gallery_dl.version
import yt_dlp.version

from . import download, output, utils, version

custom_args = output.args

log_file = output.LOG_FILE
last_line = ""
last_position = 0

log = output.initialise_logging(__name__)


async def redirect(request: Request):
    """Redirect to homepage on request."""
    return RedirectResponse(url="/gallery-dl")


async def homepage(request: Request):
    """Return homepage template response."""
    return templates.TemplateResponse(
        "index.html",
        {
            "request": request,
            "app_version": version.__version__,
            "gallery_dl_version": gallery_dl.version.__version__,
            "yt_dlp_version": yt_dlp.version.__version__,
        },
    )


async def submit_form(request: Request):
    """Process form submission data and start download in the background."""
    form_data = await request.form()

    keys = ("url", "video-opts")
    values = tuple(form_data.get(key) for key in keys)

    url, video_opts = (None if isinstance(value, UploadFile) else value for value in values)

    if not url:
        log.error("No URL provided.")

        return JSONResponse(
            {
                "success": False,
                "error": "/q called without a 'url' in form data",
            },
        )

    if not video_opts:
        video_opts = "none-selected"

    request_options = {"video-options": video_opts}

    task = BackgroundTask(download_task, url.strip(), request_options)

    log.info("Added URL to the download queue: %s", url)

    return JSONResponse(
        {
            "success": True,
            "url": url,
            "options": request_options,
        },
        background=task,
    )


def download_task(url: str, request_options: dict[str, str]):
    """Initiate download as a subprocess and log the output."""
    log_queue: Queue[dict[str, Any]] = multiprocessing.Queue()
    return_status: Queue[int] = multiprocessing.Queue()

    args = (url, request_options, log_queue, return_status, custom_args)

    process = multiprocessing.Process(target=download.run, args=args)
    process.start()

    while True:
        if log_queue.empty() and not process.is_alive():
            break
        try:
            record_dict = log_queue.get(timeout=1)
            record = output.dict_to_record(record_dict)

            if record.levelno >= output.LOG_LEVEL_MIN:
                log.handle(record)

            if "Video should already be available" in record.getMessage():
                log.warning("Terminating process as video is not available")
                process.kill()
        except queue.Empty:
            continue

    process.join()

    try:
        exit_code = return_status.get(block=False)
    except queue.Empty:
        exit_code = process.exitcode

    if exit_code == 0:
        log.info("Download process exited successfully")
    else:
        log.error("Download failed with exit code: %s", exit_code)


async def log_route(request: Request):
    """Return logs page template response."""

    async def read_log_file(file_path: str):
        log_contents = ""
        try:
            async with aiofiles.open(file_path, mode="r", encoding="utf-8") as file:
                async for line in file:
                    log_contents += line
        except FileNotFoundError:
            return "Log file not found."
        except Exception as e:
            log.debug(f"Exception: {type(e).__name__}: {e}")
            return f"An error occurred: {e}"

        return log_contents if log_contents else "No logs to display."

    logs = await read_log_file(log_file)

    return templates.TemplateResponse(
        "logs.html",
        {
            "request": request,
            "app_version": version.__version__,
            "logs": logs,
        },
    )


async def clear_logs(request: Request):
    """Clear the log file on request."""
    try:
        with open(log_file, "w") as file:
            file.write("")

        return JSONResponse(
            {
                "success": True,
                "message": "Logs successfully cleared.",
            },
            status_code=HTTP_200_OK,
        )
    except FileNotFoundError:
        return JSONResponse(
            {
                "success": False,
                "error": "Log file not found.",
            },
            status_code=HTTP_404_NOT_FOUND,
        )
    except IOError:
        return JSONResponse(
            {
                "success": False,
                "error": "An error occurred while accessing the log file.",
            },
            status_code=HTTP_500_INTERNAL_SERVER_ERROR,
        )
    except Exception as e:
        log.debug(f"Exception: {type(e).__name__}: {e}")

        return JSONResponse(
            {
                "success": False,
                "error": str(e),
            },
            status_code=HTTP_500_INTERNAL_SERVER_ERROR,
        )


async def log_stream(request: Request):
    """Stream the full contents of the log file."""

    async def file_iterator(file_path: str):
        try:
            async with aiofiles.open(file_path, mode="r", encoding="utf-8") as file:
                while True:
                    chunk = await file.read(64 * 1024)
                    if not chunk:
                        break
                    if utils.WINDOWS:
                        yield chunk.replace("\n", "\r\n")
                    else:
                        yield chunk
        except FileNotFoundError:
            yield "Log file not found."
        except Exception as e:
            log.debug(f"Exception: {type(e).__name__}: {e}")
            yield f"An error occurred: {type(e).__name__}: {e}"

    return StreamingResponse(file_iterator(log_file), media_type="text/plain")


async def log_update(websocket: WebSocket):
    """Stream log file updates over WebSocket connection."""
    global last_line, last_position

    await websocket.accept()
    log.debug(f"Accepted WebSocket connection: {websocket}")

    async with connections_lock:
        active_connections.add(websocket)
        log.debug("WebSocket added to active connections")
    try:
        async with aiofiles.open(log_file, mode="r", encoding="utf-8") as file:
            await file.seek(0, os.SEEK_END)

            async for changes in watchfiles.awatch(
                log_file,
                stop_event=shutdown_event,
                rust_timeout=100,
                yield_on_timeout=True,
            ):
                new_content = ""
                do_update_state = False

                previous_line, position = await asyncio.to_thread(
                    output.read_previous_line, log_file, last_position
                )
                if "B/s" in previous_line and previous_line != last_line:
                    new_content = previous_line
                    do_update_state = True

                new_lines = await file.read()
                if new_lines.strip():
                    new_content += new_lines

                if new_content.strip():
                    await websocket.send_text(new_content)

                    if do_update_state:
                        last_line = previous_line
                        last_position = position
                    else:
                        last_line = ""
                        last_position = 0
    except asyncio.CancelledError as e:
        log.debug(f"Exception: {type(e).__name__}")
    except WebSocketDisconnect as e:
        log.debug(f"Exception: {type(e).__name__}")
    except Exception as e:
        log.debug(f"Exception: {type(e).__name__}: {e}")
    finally:
        async with connections_lock:
            if websocket in active_connections:
                active_connections.remove(websocket)
                log.debug("WebSocket removed from active connections")


@asynccontextmanager
async def lifespan(app: Starlette):
    """Run server startup and shutdown tasks."""
    output.configure_default_loggers()

    uvicorn_log = output.get_logger("uvicorn")
    uvicorn_log.info(f"Starting {type(app).__name__} application.")

    await shutdown_override()
    try:
        yield
    except asyncio.CancelledError:
        pass
    finally:
        if utils.CONTAINER and os.path.isdir("/config"):
            if os.path.isfile(log_file) and os.path.getsize(log_file) > 0:
                dst_dir = "/config/logs"

                os.makedirs(dst_dir, exist_ok=True)

                dst = os.path.join(dst_dir, "app_" + time.strftime("%Y-%m-%d_%H-%M-%S") + ".log")
                shutil.copy2(log_file, dst)


async def shutdown_override():
    """Override uvicorn signal handlers to ensure a graceful shutdown."""
    sigint_handler = signal.getsignal(signal.SIGINT)
    sigterm_handler = signal.getsignal(signal.SIGTERM)

    def shutdown(sig: int, frame: FrameType | None = None):
        """Call shutdown handler and then original handler as a callback."""
        global shutdown_in_progress
        if shutdown_in_progress:
            return

        shutdown_in_progress = True

        event_loop = asyncio.get_event_loop()
        future = asyncio.run_coroutine_threadsafe(shutdown_handler(), event_loop)
        future.add_done_callback(lambda f: call_original_handler(sig, frame))

    def call_original_handler(sig: int, frame: FrameType | None = None):
        """Call the original signal handler for server shutdown."""
        if sig == signal.SIGINT and callable(sigint_handler):
            sigint_handler(sig, frame)
        elif sig == signal.SIGTERM and callable(sigterm_handler):
            sigterm_handler(sig, frame)

    signal.signal(signal.SIGINT, shutdown)
    signal.signal(signal.SIGTERM, shutdown)


async def shutdown_handler():
    """Initiate server shutdown."""
    if not shutdown_event.is_set():
        shutdown_event.set()
        log.debug("Set shutdown event")

    await close_connections()
    output.close_handlers()


async def close_connections():
    """Close WebSocket connections and clear the set of active connections."""
    async with connections_lock:
        log.debug(f"Active connections before closing: {len(active_connections)}")
        log.debug(f"Active tasks before closing: {len(asyncio.all_tasks())}")

        close_connections = []
        for websocket in active_connections:
            if websocket.client_state == WebSocketState.CONNECTED:
                close_connections.append(websocket.close())
                log.debug(f"Scheduled WebSocket for closure: {websocket}")

        if close_connections:
            await asyncio.gather(*close_connections)
            log.debug("Closed all WebSocket connections")

        if active_connections:
            active_connections.clear()
            log.debug("Cleared active connections")


class CSPMiddleware(BaseHTTPMiddleware):
    """Enforce Content Security Policy for all requests."""

    def __init__(self, app: ASGIApp, csp_policy: str):
        super().__init__(app)
        self.csp_policy = csp_policy

    async def dispatch(self, request, call_next):
        response = await call_next(request)
        response.headers["Content-Security-Policy"] = self.csp_policy
        return response


templates = Jinja2Templates(directory=utils.resource_path("templates"))

active_connections: set[WebSocket] = set()
connections_lock = asyncio.Lock()
shutdown_event = asyncio.Event()
shutdown_in_progress = False

routes = [
    Route("/", endpoint=redirect, methods=["GET"]),
    Route("/gallery-dl", endpoint=homepage, methods=["GET"]),
    Route("/gallery-dl/q", endpoint=submit_form, methods=["POST"]),
    Route("/gallery-dl/logs", endpoint=log_route, methods=["GET"]),
    Route("/gallery-dl/logs/clear", endpoint=clear_logs, methods=["POST"]),
    Route("/stream/logs", endpoint=log_stream, methods=["GET"]),
    WebSocketRoute("/ws/logs", endpoint=log_update),
    Mount("/static", app=StaticFiles(directory=utils.resource_path("static")), name="static"),
]

csp_policy = (
    "default-src 'self'; "
    "connect-src 'self'; "
    "form-action 'self'; "
    "manifest-src 'self'; "
    "img-src 'self' data:; "
    "script-src 'self' https://cdn.jsdelivr.net; "
    "style-src 'self' https://cdn.jsdelivr.net https://fonts.googleapis.com; "
    "font-src 'self' https://cdn.jsdelivr.net https://fonts.gstatic.com;"
)

middleware = [
    Middleware(CORSMiddleware, allow_origins=["*"], allow_methods=["POST"]),
    Middleware(CSPMiddleware, csp_policy=csp_policy),
]

app = Starlette(debug=True, routes=routes, middleware=middleware, lifespan=lifespan)



import time

# Sample string
msg = "# example message"

# Number of iterations for averaging
iterations = 10000000

# Timing slicing
start_time = time.time()
for _ in range(iterations):
    sliced_msg = msg[2:]
end_time = time.time()
slicing_time = end_time - start_time

# Timing removeprefix
start_time = time.time()
for _ in range(iterations):
    prefix_removed_msg = msg.removeprefix("# ")
end_time = time.time()
removeprefix_time = end_time - start_time

# Print results
print(f"Slicing took {slicing_time:.10f} seconds for {iterations} iterations.")
print(f"removeprefix() took {removeprefix_time:.10f} seconds for {iterations} iterations.")

input("\n" + "Press Enter to continue . . . ")



import os, sys, subprocess

prompt = os.environ.get("PROMPT", "None")

os.environ["PROMPT"] = "$P$G"

command = [sys.executable, "print_name.py"]
process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)

while True:
    if process.stdout:
        output = process.stdout.readline()

        if output == "" and process.poll() is not None:
            break

        if output:
            print(output)

status = process.wait()
print(status)

if prompt == "None":
    input("\n" + "Press Enter to continue . . . ")



# -*- coding: utf-8 -*-

from itertools import chain
from multiprocessing.queues import Queue
from typing import Any

from gallery_dl import job, exception

from . import options


def _init(custom_args: options.CustomNamespace | None):
    """Import required modules and set up basic logging.

    Ensures that parsed command-line arguments are set in the
    memory space of the current process if supplied.
    """
    global config, output, log

    if options.custom_args is None and custom_args is not None:
        options.custom_args = custom_args

    from . import config, output

    output.configure_default_loggers(is_main_process=False)

    log = output.initialise_logging(__name__)


def run(
    url: str,
    request_options: dict[str, str],
    log_queue: Queue[dict[str, Any]],
    return_status: Queue[int],
    custom_args: options.CustomNamespace | None,
):
    """Set gallery-dl configuration, set up logging and run download job."""
    _init(custom_args)

    config.clear()
    config.load()

    output.setup_logging()
    output.capture_logs(log_queue)
    output.redirect_standard_streams()

    log.info(f"Requested download with the following options: {request_options}")

    entries = config_update(request_options)

    if any(entries[0]):
        log.info(f"Added entries to the config dict: {entries[0]}")

    if any(entries[1]):
        log.info(f"Removed entries from the config dict: {entries[1]}")

    status = 0
    try:
        status = job.DownloadJob(url).run()
    except exception.GalleryDLException as e:
        status = e.code
        log.error(f"Exception: {e.__module__}.{type(e).__name__}: {e}")
    except Exception as e:
        status = -1
        log.error(f"Exception: {type(e).__name__}: {e}")
    except KeyboardInterrupt:
        pass

    output.close_handlers()

    return_status.put(status)


def config_update(request_options: dict[str, str]):
    """Update loaded configuration with request options."""
    entries_added: list[dict[str, Any] | None] = []
    entries_removed: list[Any] = []

    requested_format = request_options.get("video-options", "none-selected")

    if requested_format == "none-selected":
        return entries_added, entries_removed

    cmdline_args = config.get(["extractor", "ytdl", "cmdline-args"])
    raw_options = config.get(["extractor", "ytdl", "raw-options"])
    postprocessors = config.get(["postprocessors"], conf=raw_options)

    if requested_format == "download-video":
        entries_removed.extend(
            chain(
                config.remove(cmdline_args, item="--extract-audio"),
                config.remove(cmdline_args, item="-x"),
                config.remove(raw_options, key="writethumbnail", value=False),
                config.remove(postprocessors, key="key", value="FFmpegExtractAudio"),
            )
        )

    if requested_format == "extract-audio":
        entries_added.extend(
            chain(
                config.add(
                    {
                        "extractor": {
                            "ytdl": {
                                "cmdline-args": [
                                    "--extract-audio",
                                ]
                            }
                        }
                    }
                )[1],
                config.add(
                    {
                        "extractor": {
                            "ytdl": {
                                "raw-options": {
                                    "writethumbnail": False,
                                    "postprocessors": [
                                        {
                                            "key": "FFmpegExtractAudio",
                                            "preferredcodec": "best",
                                            "preferredquality": 320,
                                        }
                                    ],
                                }
                            }
                        }
                    }
                )[1],
            )
        )

        entries_removed.extend(
            config.remove(cmdline_args, item="--merge-output-format", value="any")
        )

    return entries_added, entries_removed



# -*- coding: utf-8 -*-

import uvicorn

from . import options


def main(
    app: str = "gallery_dl_server.server:app",
    args: options.CustomNamespace | None = None,
    is_main_module: bool = False,
):
    """Main entry point for gallery-dl-server."""
    if args is None:
        args = options.parse_args(is_main_module)

    kwargs = {
        "host": args.host,
        "port": args.port,
        "log_level": args.server_log_level,
        "access_log": args.access_log,
    }

    try:
        uvicorn.run(app, **kwargs)
    except KeyboardInterrupt:
        pass



# -*- coding: utf-8 -*-

import os
import sys

import asyncio
import io
import logging
import pickle
import queue
import re
import threading

from mmap import mmap, ACCESS_READ, ACCESS_WRITE
from multiprocessing.queues import Queue
from typing import TextIO, Any

from gallery_dl import output, job

from . import options, utils

args = options.custom_args

if args is None:
    args = options.get_default_args()

log_dir = args.log_dir
log_level = args.log_level
server_log_level = args.server_log_level
access_log = args.access_log

if server_log_level == "trace":
    server_log_level = "debug"

LOG_FILE = utils.get_log_file_path(log_dir)
LOG_LEVEL: int = getattr(logging, server_log_level.upper())
LOG_LEVEL_DL: int = getattr(logging, log_level.upper())
LOG_LEVEL_MIN = min(LOG_LEVEL, LOG_LEVEL_DL)
LOG_FORMAT = "%(asctime)s [%(levelname)s] %(message)s"
LOG_FORMAT_DEBUG = "%(asctime)s [%(name)s] [%(filename)s:%(lineno)d] [%(levelname)s] %(message)s"
LOG_FORMAT_DATE = "%Y-%m-%d %H:%M:%S"
LOG_SEPARATOR = "/sep/"


def initialise_logging(
    name=utils.get_package_name(),
    stream: TextIO | Any = sys.__stdout__,
    file=LOG_FILE,
    level=LOG_LEVEL,
):
    """Set up basic logging functionality for gallery-dl-server."""
    logger = AsyncLogger(name, level)
    logger.propagate = False

    if not logger.hasHandlers():
        formatter = CustomFormatter(LOG_FORMAT, LOG_FORMAT_DATE)

        handler_console = setup_stream_handler(stream, formatter)
        logger.addHandler(handler_console)

        if file:
            handler_file = setup_file_handler(file, formatter)
            logger.addHandler(handler_file)

    return logger


class AsyncLogger(logging.Logger):
    """Custom logger for async logging and logging multi-line messages."""

    def __init__(self, name: str, level=logging.NOTSET):
        super().__init__(name, level)

    def log_multiline(self, level: int, msg: str):
        """Log each line of a multi-line message separately."""
        for line in msg.split("\n"):
            if line.strip():
                self.log(level, line)

    def handle(self, record):
        """Override handle method to call the async version."""
        try:
            event_loop = asyncio.get_running_loop()
            event_loop.create_task(self.handle_async(record))
        except RuntimeError:
            asyncio.run(self.handle_async(record))

    async def handle_async(self, record: logging.LogRecord):
        """Perform asynchronous handling of log records."""
        super().handle(record)


class CustomFormatter(logging.Formatter):
    """Custom formatter for log messages."""

    def format(self, record):
        record.levelname = record.levelname.lower()

        msg = super().format(record)
        return remove_ansi_escape_sequences(msg).rstrip()


def remove_ansi_escape_sequences(text: str, /):
    """Remove ANSI escape sequences from the given text."""
    ansi_escape_pattern = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
    return ansi_escape_pattern.sub("", text)


def setup_stream_handler(stream: TextIO | Any, formatter: logging.Formatter):
    """Set up a console handler for logging."""
    stream_handler = logging.StreamHandler(stream)
    stream_handler.setFormatter(formatter)
    register_handler(stream_handler)

    return stream_handler


def setup_file_handler(file: str, formatter: logging.Formatter):
    """Set up a file handler for logging."""
    os.makedirs(os.path.dirname(file), exist_ok=True)

    file_handler = logging.FileHandler(file, mode="a", encoding="utf-8", delay=False)
    file_handler.setFormatter(formatter)
    register_handler(file_handler)

    return file_handler


def get_logger(name: str | None = None):
    """Return a logger with the specified name."""
    return logging.getLogger(name)


def get_blank_logger(name="blank", stream=sys.stdout, level=logging.INFO):
    """Return a basic logger with no formatter."""
    logger = AsyncLogger(name, level)
    logger.propagate = False

    if not logger.hasHandlers():
        handler = logging.StreamHandler(stream)
        logger.addHandler(handler)
        register_handler(handler)

    return logger


def setup_logging(level=LOG_LEVEL_DL):
    """Set up gallery-dl logging."""
    logger = output.initialize_logging(level)

    output.configure_standard_streams()
    output.configure_logging(level)

    handler = output.setup_logging_handler("unsupportedfile", fmt="{message}")

    if handler:
        ulog = logging.getLogger("unsupported")
        ulog.addHandler(handler)
        ulog.propagate = False

        register_handler(handler)
        setattr(job.Job, "ulog", ulog)

    return logger


def capture_logs(log_queue: Queue[dict[str, Any]]):
    """Send logs that reach the root logger to a queue."""
    root = logging.getLogger()
    queue_handler = QueueHandler(log_queue)

    if root.handlers:
        existing_handler = root.handlers[0]
        queue_handler.setFormatter(existing_handler.formatter)

        for handler in root.handlers[:]:
            if isinstance(handler, logging.StreamHandler):
                handler.close()
                root.removeHandler(handler)

    root.addHandler(queue_handler)
    register_handler(handler)


class QueueHandler(logging.Handler):
    """Custom logging handler that sends log messages to a queue."""

    def __init__(self, queue: Queue[dict[str, Any]]):
        super().__init__()
        self.queue = queue

    def emit(self, record: logging.LogRecord):
        record.msg = self.format(record).strip()
        record.args = ()
        record_dict = record_to_dict(record)

        self.queue.put(record_dict)


def record_to_dict(record: logging.LogRecord):
    """Convert a log record into a dictionary."""
    record_dict = record.__dict__.copy()
    record_dict["level"] = record.levelno

    sanitise_dict(record_dict)

    return record_dict


def sanitise_dict(record_dict: dict[str, Any]):
    """Remove non-serialisable values from a dictionary."""
    keys_to_remove = []

    for key, value in record_dict.items():
        if not is_serialisable(value):
            keys_to_remove.append(key)

    for key in keys_to_remove:
        record_dict.pop(key)


def is_serialisable(value: Any):
    """Check if a value can be serialised."""
    try:
        pickle.dumps(value)
        return True
    except Exception:
        return False


def dict_to_record(record_dict: dict[str, Any]):
    """Convert a dictionary back into a log record."""
    return logging.LogRecord(**record_dict)


def stdout_write(s: str, /):
    """Write directly to stdout."""
    sys.stdout.write(s + "\n")
    sys.stdout.flush()


def stderr_write(s: str, /):
    """Write directly to stderr."""
    sys.stderr.write(s + "\n")
    sys.stderr.flush()


def redirect_standard_streams(level=logging.INFO):
    """Redirect stdout and stderr streams and log at level."""
    logger_writer = LoggerWriter(level)

    setattr(sys, "stdout", logger_writer)
    setattr(sys, "stderr", logger_writer)


class LoggerWriter:
    """Log writes to stdout and stderr."""

    def __init__(self, level=logging.INFO):
        self.level = level
        self.stream = sys.__stdout__
        self.logger = initialise_logging(type(self).__name__)

        for handler in self.logger.handlers[:]:
            handler.close()
            self.logger.removeHandler(handler)

        self.stream_handler = ConsoleProgress(self.stream)
        self.file_handler = FileProgress()

        self.logger.addHandler(self.stream_handler)
        self.logger.addHandler(self.file_handler)

        register_handler(self.stream_handler, self.file_handler)

    def write(self, msg: str, /):
        """Prepare and log messages."""
        msg = msg.strip()

        if not msg:
            return

        if msg.startswith("* "):
            msg = f"Download successful: {msg[2:]}"

        if msg.startswith("# "):
            msg = f"File already exists or is in download archive: {msg[2:]}"
            return self.logger.log(logging.WARNING, msg)

        self.logger.log(self.level, msg)

    def flush(self):
        pass


class NullStream(io.TextIOBase):
    """Suppress writes to stdout or stderr."""

    def write(self, _: str, /):
        pass

    def flush(self):
        pass


class ConsoleProgress(logging.StreamHandler):
    """Custom stream handler that can handle progress updates."""

    def __init__(self, stream: TextIO | Any = NullStream()):
        super().__init__(stream)
        self.stream: TextIO | Any
        self.formatter = CustomFormatter(LOG_FORMAT, LOG_FORMAT_DATE)
        self.last_msg = ""

    def emit(self, record):
        """Override emit method to write final progress update."""
        msg = self.format(record)

        if "B/s" not in msg and "B/s" in self.last_msg:
            self.stream.write(self.last_msg + "\n")
            self.stream.flush()

        if "B/s" not in msg:
            super().emit(record)

        self.last_msg = msg


class FileProgress(logging.FileHandler):
    """Custom file handler that can handle progress updates asynchronously."""

    def __init__(self, filename=LOG_FILE, mode="a", encoding="utf-8", delay=False):
        super().__init__(filename, mode, encoding, delay)
        self.formatter = CustomFormatter(LOG_FORMAT, LOG_FORMAT_DATE)
        self.last_pos = 0
        self.queue = queue.Queue()
        self.thread = threading.Thread(target=self.process_queue, daemon=True)
        self.thread_lock = threading.Lock()
        self.thread.start()

    def emit(self, record):
        """Override emit method to queue progress messages."""
        msg = self.format(record)

        if "B/s" in msg:
            self.queue.put(msg)
        else:
            with self.thread_lock:
                super().emit(record)

    def process_queue(self):
        """Process logging queue in a separate thread."""
        while True:
            msg = self.queue.get()
            if msg is None:
                break
            with self.thread_lock:
                self.write_to_file(msg)

    def write_to_file(self, msg: str):
        """Write download progress to the memory-mapped log file."""
        with open(self.baseFilename, "r+b") as file:
            with mmap(file.fileno(), 0, access=ACCESS_WRITE) as mm:
                self.write_progress(mm, msg)

    def write_progress(self, mmap: mmap, msg: str):
        """Write download progress updates to the same line."""
        if self.last_pos == 0:
            self.last_pos = mmap.rfind(b"\n") + 1

        mmap.seek(self.last_pos)

        new_msg = msg.encode("utf-8")
        new_msg_length = len(new_msg)
        target_msg_length = 54

        padding = b""
        if new_msg_length < target_msg_length:
            padding = b" " * (target_msg_length - new_msg_length)

        terminator = b"\n"
        new_size = self.last_pos + max(new_msg_length, target_msg_length) + 1

        if utils.WINDOWS:
            terminator = b"\r\n"
            new_size += 1

        if new_size > mmap.size():
            mmap.resize(new_size)

        mmap.write(new_msg + padding + terminator)
        mmap.flush()

    def close(self):
        """Stop logging thread and close handler."""
        self.queue.put(None)
        self.thread.join()
        super().close()


def read_previous_line(file_path: str, last_position: int):
    """Return the previous line of a file or from a given position."""
    previous_line = ""
    position = 0

    if os.path.getsize(file_path) == 0:
        return previous_line, position

    with open(file_path, "rb") as file:
        with mmap(file.fileno(), 0, access=ACCESS_READ) as mm:
            if last_position == 0:
                position = mm.rfind(b"\n", 0, mm.size() - 2) + 1
                if position == 0:
                    return previous_line, position
            else:
                position = last_position

            next_line = mm.find(b"\n", position) + 1
            previous_line = mm[position:next_line].decode("utf-8")

    return previous_line, position


class StringLogger:
    """Add StringHandler to the root logger and get logs."""

    def __init__(self, level=LOG_LEVEL):
        self.root = logging.getLogger()

        self.handler = StringHandler()
        self.handler.setLevel(level)

        self.root.addHandler(self.handler)

    def get_logs(self):
        """Return logs captured by the handler."""
        return self.handler.get_logs()

    def close(self):
        """Close and remove the handler from the root logger."""
        self.handler.close()
        self.root.removeHandler(self.handler)


class StringHandler(logging.Handler):
    """Capture log messages and write them to a string object."""

    def __init__(self):
        super().__init__()
        self.buffer = io.StringIO()
        self.terminator = LOG_SEPARATOR

    def emit(self, record):
        msg = self.format(record)

        if msg.strip():
            self.buffer.write(msg + self.terminator)

    def get_logs(self):
        """Retrieve logs from the buffer."""
        logs = self.buffer.getvalue()
        return logs.rstrip(self.terminator)

    def close(self):
        """Close the log buffer and handler."""
        self.buffer.close()
        super().close()


def last_line(msg: str, string: str, case_sensitive=True):
    """Get the last line containing a string in a message."""
    lines = msg.split("\n")

    for line in reversed(lines):
        if (string in line) if case_sensitive else (string.lower() in line.lower()):
            return line

    return None


def join(_list: list[str]):
    """Join a list of strings with commas and single quotes."""
    formatted_list = []

    for item in _list:
        formatted_list.append(f"'{item}'")

    return ", ".join(formatted_list)


def configure_default_loggers(is_main_process=True):
    """Configure default logging behaviour."""
    asyncio_log = logging.getLogger("asyncio")
    asyncio_log.setLevel(logging.CRITICAL)

    if is_main_process:
        uvicorn_error = logging.getLogger("uvicorn.error")

        if not access_log:
            uvicorn_error.addFilter(WebSocketFilter())


class WebSocketFilter(logging.Filter):
    """Filter out WebSocket connection logs."""

    def filter(self, record):
        msg = record.getMessage()
        strings = ["WebSocket /ws/logs", "connection open", "connection closed"]
        return not any(string in msg for string in strings)


def register_handler(*handlers: logging.Handler):
    """Add handler to the logging manager."""
    logging_manager = LoggingManager()

    for handler in handlers:
        logging_manager.add_handler(handler)


def close_handlers():
    """Close all registered logging handlers."""
    logging_manager = LoggingManager()
    logging_manager.close_all()


class LoggingManager:
    """Store logging handler instances for closure."""

    _instance: "LoggingManager | None" = None
    logger: "BasicLogger"

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(LoggingManager, cls).__new__(cls)
            cls._instance.logger = BasicLogger()
            cls._instance.logger.debug("Created new instance of LoggingManager.")

        return cls._instance

    def __init__(self):
        if not hasattr(self, "handlers"):
            self.handlers: list[logging.Handler] = []
            self.logger.debug("Initialized handlers list.")

    def add_handler(self, handler: logging.Handler):
        """Add handler to list of handlers."""
        self.handlers.append(handler)
        self.logger.debug(f"Added handler: {handler}")

    def close_all(self):
        """Close all logging handlers and clear the list."""
        for handler in self.handlers[:]:
            handler.close()
            self.logger.debug(f"Closed handler: {handler}")

        self.handlers.clear()
        self.logger.debug("Cleared handlers list.")
        self.logger.close()


class BasicLogger(logging.Logger):
    """Basic logger for debugging."""

    def __init__(self):
        super().__init__(name=type(self).__name__, level=LOG_LEVEL)
        self.addHandler(BasicHandler())
        self.propagate = False

    def close(self):
        """Close and remove all handlers from the logger."""
        for handler in self.handlers[:]:
            handler.close()
            self.removeHandler(handler)


class BasicHandler(logging.Handler):
    """Basic handler for debugging."""

    def __init__(self):
        super().__init__()
        self.formatter = CustomFormatter(LOG_FORMAT, LOG_FORMAT_DATE)

    def emit(self, record):
        sys.stdout.write(self.format(record) + "\n")
